import os

from charmhelpers.core import hookenv, host, templating, unitdata
from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.fetch import apt_purge
from charms import apt, reactive


@reactive.hook('upgrade-charm')
def upgrade_charm():
    hookenv.status_set('maintenance', 'forcing reconfiguration on upgrade-charm')
    reactive.clear_flag('jenkins-slave.active')
    reactive.clear_flag('jenkins-slave.installed')


@reactive.when_not('jenkins-slave.installed')
def install():
    hookenv.status_set('maintenance', 'installing jenkins-slave')
    reactive.clear_flag('jenkins-slave.active')

    config = hookenv.config()

    hookenv.log('Adding jenkins-slave dependencies to be installed')
    packages = ['wget', 'default-jre-headless']

    # Install extra packages needed by the slave.
    tools = config.get('tools')
    if tools:
        hookenv.log('Adding jenkins-slave additional tools to be installed: {}'.format(tools))
        for package in tools.split():
            packages.append(package)
    apt.queue_install(packages)
    if not apt.install_queued():
        return  # apt layer already set blocked state.

    # Get rid of the legacy jenkins-slave package, including config files.
    hookenv.log('Removing the old jenkins-slave package... (obsoleted by this charm)')
    apt_purge(['jenkins-slave'])

    # Create jenkins user if it doesn't exist.
    if host.user_exists('jenkins'):
        hookenv.log('Installing jenkins-slave (user account already exists)...')
    else:
        hookenv.log('Installing jenkins-slave (user account)...')
        host.adduser(username='jenkins', system_user=True, home_dir='/var/lib/jenkins')

    # And ensure required directories exist and are set up.
    hookenv.log('Installing jenkins-slave (directories)...')
    host.mkdir('/var/lib/jenkins', owner='jenkins', group='jenkins')
    host.mkdir('/var/log/jenkins', owner='jenkins', group='jenkins')

    hookenv.log('Installing jenkins-slave (common files)...')
    write_default_conf()
    file_to_units('files/download-slave.sh', '/usr/local/sbin/download-slave.sh')
    file_to_units('files/jenkins-slave-logrotate-config', '/etc/logrotate.d/jenkins-slave')

    if host.lsb_release()['DISTRIB_CODENAME'] == 'trusty':
        hookenv.log('Installing jenkins-slave (upstart job)...')
        file_to_units('files/jenkins-slave-upstart-config', '/etc/init/jenkins-slave.conf')
    else:
        hookenv.log('Installing jenkins-slave (system unit)...')
        file_to_units('files/jenkins-slave-systemd-config', '/lib/systemd/system/jenkins-slave.service')
        host.service('enable', 'jenkins-slave')

    hookenv.log('Installing jenkins-slave... done.')
    reactive.clear_flag('jenkins-slave.blocked')
    reactive.clear_flag('jenkins-slave.configured')
    reactive.set_flag('jenkins-slave.installed')


@reactive.when('config.changed')
def config_changed():
    reactive.clear_flag('jenkins-slave.blocked')
    reactive.clear_flag('jenkins-slave.configured')
    reactive.clear_flag('nagios-nrpe.configured')


@reactive.when('jenkins-slave.installed')
@reactive.when_not('jenkins-slave.configured')
@reactive.when_not('jenkins-slave.blocked')
def configure_jenkins_slave():
    hookenv.status_set('maintenance', 'configuring jenkins-slave')
    reactive.clear_flag('jenkins-slave.active')

    config = hookenv.config()
    kv = unitdata.kv()

    if config.get('master_url'):
        hookenv.log("Using 'master_url' to configure the slave.")
        write_default_conf(config.get('master_url'))
    elif kv.get('url'):
        hookenv.log("Using url from relation as 'master_url'")
        write_default_conf(kv.get('url'))
    else:
        hookenv.log("No 'master_url' set; not configuring slave at this time.")
        hookenv.status_set('blocked', "requires either slave relation or 'master_url'")
        reactive.set_flag('jenkins-slave.blocked')
        return

    file_to_units('files/jenkins-slave-sudoers', '/etc/sudoers.d/jenkins', perms=0o440)

    reactive.clear_flag('nagios-nrpe.configured')
    reactive.set_flag('jenkins-slave.configured')


@reactive.when('jenkins-slave.blocked')
def blocked_on_jenkins_url():
    reactive.clear_flag('jenkins-slave.active')


@reactive.when('jenkins-slave.configured')
@reactive.when('nrpe-external-master.available')
@reactive.when('jenkins-slave.active')
@reactive.when_not('nagios-nrpe.configured')
def configure_nagios(nagios):
    hookenv.status_set('maintenance', 'setting up NRPE checks')

    # Use charmhelpers.contrib.charmsupport's nrpe to determine hostname
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname, primary=True)

    cmd = '/usr/lib/nagios/plugins/check_procs -c 1:1 -a slave.jar'
    nrpe_setup.add_check('jenkins_slave_ps', 'Jenkins Slave Process', cmd)

    nrpe_setup.write()
    reactive.set_flag('nagios-nrpe.configured')


@reactive.when('jenkins-slave.configured')
@reactive.when_not('jenkins-slave.active')
def set_active():
    # Startup the jenkins-slave service. This is called in the
    # install, config-changed and upgrade-charm paths. It needs to be
    # tolerant of a running or stopped jenkins-slave.
    if host.service_running('jenkins-slave'):
        hookenv.log('Restarting jenkins-slave...')
        host.service_restart('jenkins-slave')
    else:
        hookenv.log('Starting jenkins-slave...')
        host.service_start('jenkins-slave')

    hookenv.status_set('active', 'ready')
    reactive.set_flag('jenkins-slave.active')


# We can't use interface:jenkins-slave yet as it's not implemented.
@reactive.hook('slave-relation-joined', 'slave-relation-changed')
def slave_relation_changed():
    reactive.set_flag('slave-relation.available')
    reactive.clear_flag('jenkins-slave.blocked')
    reactive.clear_flag('jenkins-slave.configured')
    reactive.clear_flag('slave-relation.configured')


@reactive.hook('slave-relation-departed', 'slave-relation-broken')
def slave_relation_removed():
    kv = unitdata.kv()
    kv.set('url', None)
    reactive.clear_flag('slave-relation.available')


@reactive.when('slave-relation.available')
@reactive.when_not('slave-relation.configured')
def slave_relation():
    hookenv.status_set('maintenance', 'setting up jenkins via slave relation')
    config = hookenv.config()
    kv = unitdata.kv()

    if config.get('master_url'):
        hookenv.log("Config option 'master_url' is set. Can't use slave relation.")
        reactive.set_flag('slave-relation.configured')
        return

    url = hookenv.relation_get('url')
    if url:
        kv.set('url', url)
        write_default_conf(url)
    else:
        hookenv.log("Master hasn't exported its url yet, exiting...")
        return

    reactive.clear_flag('jenkins-slave.active')

    # Set the slave hostname to match the juju unit
    # in the jenkins master instance
    slave_host = hookenv.local_unit().replace('/', '-')
    slave_address = hookenv.unit_private_ip()
    noexecutors = os.cpu_count()
    config_labels = config.get('labels')

    if config_labels:
        labels = config_labels
    else:
        labels = os.uname()[4]

    # Set all relations
    hookenv.relation_set(executors=noexecutors)
    hookenv.relation_set(labels=labels)
    hookenv.relation_set(slavehost=slave_host)
    hookenv.relation_set(slaveaddress=slave_address)
    reactive.set_flag('slave-relation.configured')


def file_to_units(local_path, unit_path, perms=None, owner='root', group='root'):
    """ copy a file from the charm onto our unit(s) """
    file_perms = perms
    if not perms:
        # Let's try manually work it out
        if local_path[-3:] == '.py' or local_path[-3:] == '.sh':
            file_perms = 0o755
        else:
            file_perms = 0o644

    with open(local_path, 'r') as fh:
        host.write_file(
            path=unit_path,
            content=fh.read().encode(),
            owner=owner,
            group=group,
            perms=file_perms,
        )


def write_default_conf(master_url=None, owner='root', group='root',
                       conf_path='/etc/default/jenkins-slave'):
    templates_dir = os.path.join(hookenv.charm_dir(), 'templates')
    slave_host = hookenv.local_unit().replace('/', '-')
    templating.render('jenkins-slave-default', conf_path,
                      {'master_url': master_url, 'slave_host': slave_host},
                      owner=owner, group=group, perms=0o444,
                      templates_dir=templates_dir)
