import grp
import os
import pwd
import shutil
import sys
import tempfile
import unittest
from unittest import mock

sys.modules['charms.apt'] = mock.MagicMock()
from charms import apt  # NOQA: E402

# Add path to where our reactive layer lives and import.
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
from reactive.jenkins_slave import (
    config_changed,
    install,
    configure_jenkins_slave,
    file_to_units,
    write_default_conf,
)  # NOQA: E402


INITIAL_CONF = 'tests/unit/files/jenkins-slave-default'


class TestSetDefaultConf(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix='charm-unittests-')
        temp_file = tempfile.NamedTemporaryFile(delete=False, dir=self.tmpdir)
        with open(INITIAL_CONF, 'rb') as f:
            conf = f.read().decode('utf-8')
        temp_file.write(conf.encode())
        temp_file.close()
        self.conf_file = temp_file.name
        self.user = pwd.getpwuid(os.getuid()).pw_name
        self.group = grp.getgrgid(os.getgid()).gr_name
        self.charm_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

        # charmhelpers is getting difficult to test against, as it writes
        # to system directories even for things that should be idempotent,
        # like accessing config options.
        patcher = mock.patch('charmhelpers.core.hookenv.charm_dir')
        self.mock_charm_dir = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_charm_dir.return_value = self.charm_dir

        patcher = mock.patch('charmhelpers.core.hookenv.config')
        self.mock_config = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_config.return_value = {'tools': []}

        patcher = mock.patch('charmhelpers.core.hookenv.local_unit')
        self.mock_local_unit = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_local_unit.return_value = 'mock-jenkins-slave/0'

        patcher = mock.patch('charmhelpers.core.hookenv.log')
        self.mock_log = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_log.return_value = ''

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    @mock.patch('charms.reactive.clear_flag')
    def test_hook_config_changed(self, clear_flag):
        config_changed()
        expected = [mock.call('jenkins-slave.blocked'),
                    mock.call('jenkins-slave.configured'),
                    mock.call('nagios-nrpe.configured')]
        self.assertEqual(clear_flag.call_args_list, expected)

    @mock.patch('charmhelpers.core.host.adduser')
    @mock.patch('charmhelpers.core.host.mkdir')
    @mock.patch('charmhelpers.core.host.service')
    @mock.patch('reactive.jenkins_slave.apt_purge')
    @mock.patch('reactive.jenkins_slave.file_to_units')
    @mock.patch('reactive.jenkins_slave.write_default_conf')
    def test_hook_install(self, write_default_conf, file_to_units, apt_purge, service, mkdir, adduser):
        install()
        expected = [mock.call(home_dir='/var/lib/jenkins', system_user=True, username='jenkins')]
        self.assertEqual(adduser.call_args_list, expected)
        expected = [mock.call('/var/lib/jenkins', group='jenkins', owner='jenkins'),
                    mock.call('/var/log/jenkins', group='jenkins', owner='jenkins')]
        self.assertEqual(mkdir.call_args_list, expected)
        self.assertEqual(service.call_args_list, [mock.call('enable', 'jenkins-slave')])
        self.assertEqual(apt_purge.call_args_list, [mock.call(['jenkins-slave'])])
        expected = [mock.call('files/download-slave.sh', '/usr/local/sbin/download-slave.sh'),
                    mock.call('files/jenkins-slave-logrotate-config', '/etc/logrotate.d/jenkins-slave'),
                    mock.call('files/jenkins-slave-systemd-config', '/lib/systemd/system/jenkins-slave.service')]
        self.assertEqual(file_to_units.call_args_list, expected)
        self.assertEqual(write_default_conf.call_args_list, [mock.call()])
        expected = [mock.call.queue_install(['wget', 'default-jre-headless']),
                    mock.call.install_queued()]
        self.assertEqual(apt.method_calls, expected)

    @mock.patch('charms.reactive.clear_flag')
    @mock.patch('charms.reactive.set_flag')
    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.unitdata.kv')
    @mock.patch('reactive.jenkins_slave.write_default_conf')
    @mock.patch('reactive.jenkins_slave.file_to_units')
    def test_configure_jenkins_slave_no_url(self, file_to_units, write_default_conf, unitdata_kv, config,
                                            set_flag, clear_flag):
        config.return_value = {}
        unitdata_kv.return_value = {}
        configure_jenkins_slave()
        self.assertEqual(write_default_conf.call_args_list, [])
        self.assertEqual(set_flag.call_args_list, [mock.call('jenkins-slave.blocked')])
        self.assertEqual(clear_flag.call_args_list, [mock.call('jenkins-slave.active')])

    @mock.patch('charms.reactive.clear_flag')
    @mock.patch('charms.reactive.set_flag')
    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.unitdata.kv')
    @mock.patch('reactive.jenkins_slave.write_default_conf')
    @mock.patch('reactive.jenkins_slave.file_to_units')
    def test_configure_jenkins_slave_master_url(self, file_to_units, write_default_conf, unitdata_kv, config,
                                                set_flag, clear_flag):
        config.return_value = {'master_url': 'http://10.1.1.1:8080'}
        unitdata_kv.return_value = {}
        configure_jenkins_slave()
        self.assertEqual(write_default_conf.call_args_list, [mock.call('http://10.1.1.1:8080')])
        self.assertEqual(set_flag.call_args_list, [mock.call('jenkins-slave.configured')])
        expected = [mock.call('jenkins-slave.active'), mock.call('nagios-nrpe.configured')]
        self.assertEqual(clear_flag.call_args_list, expected)

    @mock.patch('charms.reactive.clear_flag')
    @mock.patch('charms.reactive.set_flag')
    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.unitdata.kv')
    @mock.patch('reactive.jenkins_slave.write_default_conf')
    @mock.patch('reactive.jenkins_slave.file_to_units')
    def test_configure_jenkins_slave_relation_url(self, file_to_units, write_default_conf, unitdata_kv, config,
                                                  set_flag, clear_flag):
        config.return_value = {}
        unitdata_kv.return_value = {'url': 'http://10.22.22.22:8080'}
        configure_jenkins_slave()
        print(write_default_conf.call_args_list)
        self.assertEqual(write_default_conf.call_args_list, [mock.call('http://10.22.22.22:8080')])
        self.assertEqual(set_flag.call_args_list, [mock.call('jenkins-slave.configured')])
        expected = [mock.call('jenkins-slave.active'), mock.call('nagios-nrpe.configured')]
        self.assertEqual(clear_flag.call_args_list, expected)

    def test_write_default_conf_update(self):
        write_default_conf('http://10.1.1.1:8080', self.user, self.group, self.conf_file)
        self.assertTrue(conf_match(self.conf_file, 'JENKINS_URL', 'http://10.1.1.1:8080'))

    def test_write_default_conf_reset(self):
        write_default_conf(None, self.user, self.group, self.conf_file)
        self.assertTrue(conf_match(self.conf_file, '#JENKINS_URL', ''))

    def test_file_to_units_executable_sh(self):
        source = os.path.join(self.charm_dir, 'files/download-slave.sh')
        dest = os.path.join(self.tmpdir, os.path.basename(source))
        file_to_units(source, dest, owner=self.user, group=self.group)
        with open(dest, 'rb') as fh:
            want = fh.read().decode('utf-8')
        self.assertTrue(conf_equals(source, want))
        self.assertEqual(pwd.getpwuid(os.stat(dest).st_uid).pw_name, self.user)
        self.assertEqual(grp.getgrgid(os.stat(dest).st_gid).gr_name, self.group)
        self.assertTrue(os.access(dest, os.X_OK))

    def test_file_to_units_executable_py(self):
        source = os.path.join(self.charm_dir, 'tests/unit/files/somefile.py')
        dest = os.path.join(self.tmpdir, os.path.basename(source))
        file_to_units(source, dest, owner=self.user, group=self.group)
        with open(dest, 'rb') as fh:
            want = fh.read().decode('utf-8')
        self.assertTrue(conf_equals(source, want))
        self.assertEqual(pwd.getpwuid(os.stat(dest).st_uid).pw_name, self.user)
        self.assertEqual(grp.getgrgid(os.stat(dest).st_gid).gr_name, self.group)
        self.assertTrue(os.access(dest, os.X_OK))

    def test_file_to_units_non_executable(self):
        source = os.path.join(self.charm_dir, 'files/jenkins-slave-logrotate-config')
        dest = os.path.join(self.tmpdir, os.path.basename(source))
        file_to_units(source, dest, owner=self.user, group=self.group)
        with open(dest, 'rb') as fh:
            want = fh.read().decode('utf-8')
        self.assertTrue(conf_equals(dest, want))
        self.assertEqual(pwd.getpwuid(os.stat(dest).st_uid).pw_name, self.user)
        self.assertEqual(grp.getgrgid(os.stat(dest).st_gid).gr_name, self.group)
        self.assertFalse(os.access(dest, os.X_OK))

    def test_file_to_units_non_executable_x_on_disk(self):
        source = os.path.join(self.charm_dir, 'tests/unit/files/somefile')
        dest = os.path.join(self.tmpdir, os.path.basename(source))
        file_to_units(source, dest, owner=self.user, group=self.group)
        with open(dest, 'rb') as fh:
            want = fh.read().decode('utf-8')
        self.assertTrue(conf_equals(source, want))
        self.assertEqual(pwd.getpwuid(os.stat(dest).st_uid).pw_name, self.user)
        self.assertEqual(grp.getgrgid(os.stat(dest).st_gid).gr_name, self.group)
        self.assertFalse(os.access(dest, os.X_OK))


def conf_equals(conf_file, want):
    with open(conf_file, 'rb') as conf:
        got = conf.read().decode('utf-8')
        if got == want:
            return True
        print('{}\n != \n{}'.format(got, want))
    return False


def conf_match(conf_file, key, value):
    with open(conf_file, 'rb') as conf:
        for line in conf.readlines():
            line = line.decode('utf-8').rstrip('\n')
            if line == '{}="{}"'.format(key, value):
                return True
    return False


if __name__ == '__main__':
    unittest.main()
